#include <iostream>
#include <fstream>
#include <set>
#include <gmpxx.h>
#include "clock.hpp"

/**
 * See projecteuler.com for details.
 * Problem 1:
 * If we list all the natural numbers below 10 that are multiples of 3
 * or 5, we get 3, 5, 6 and 9. The sum of these multiples is 23.  Find
 * the sum of all the multiples of 3 or 5 below 1000. 
 */

const unsigned long MAX = 1000;

unsigned long method1(const unsigned long max) {
  unsigned long sum=0;
  for (unsigned long i=0; i<max; i++) {
	if ((i % 3) == 0 || ((i % 5) == 0))
	  sum += i;
  }
  return sum;
}

unsigned long method2(const unsigned long max) {
  unsigned long sum=0;
  std::set<unsigned long> torrent;
  for (unsigned long i=0; i<max; i=i+3) {
	torrent.insert(i);
  }
  for (unsigned long i=0; i<max; i=i+5) {
	torrent.insert(i);
  }
  std::set<unsigned long>::iterator it;
  for (it=torrent.begin(); it != torrent.end(); it++) {
	sum += (*it);
  }
  return sum;
}

mpz_class method3(const unsigned long max) {
  mpz_class sum(0);
  for (unsigned long i=0; i<max; i++) {
	if ((i % 3) == 0 || ((i % 5) == 0))
	  sum += i;
  }
  return sum;
}

mpz_class method4(const unsigned long max) {
  mpz_class sum(0);
  mpz_class bigint_max(max);
  for (mpz_class i=0; i<bigint_max; i++) {
	if ((i % 3) == 0 || ((i % 5) == 0))
	  sum += i;
  }
  return sum;
}

unsigned long long method5(const unsigned long long max) {
  unsigned long long sum=0;
  for (unsigned long long i=0; i<max; i++) {
	if ((i % 3) == 0 || ((i % 5) == 0))
	  sum += i;
  }
  return sum;
}

int main(void) {
  pe::Timer timer;
  unsigned long sum=0;
  mpz_class sum2;
  unsigned long long sum3;
  std::ofstream methods("methods.txt", std::ofstream::out);
  std::ofstream scaling("scaling.txt", std::ofstream::out);

  // Writing headers
  std::cout << "Project Euler - Problem 1" << std::endl;
  methods << "Method \t Runtime" << std::endl;
  scaling << "Maximum\tMethod4\tMethod5" << std::endl;

  // Comparing the methods for MAX iterations
  timer.start();
  sum=method1(MAX);
  timer.finish();
  std::cout << "Method 1: "<< sum << " -> "<< timer.renderToSecond() << " s." << std::endl;
  methods << "Method 1" << "\t" << timer.renderToSecond() << std::endl;
  
  timer.start();
  sum=method2(MAX);
  timer.finish();
  std::cout << "Method 2: "<< sum << " -> "<< timer.renderToSecond() << " s." << std::endl;
  methods << "Method 2" << "\t" << timer.renderToSecond() << std::endl;

  timer.start();
  sum2=method3(MAX);
  timer.finish();
  std::cout << "Method 3: "<< sum << " -> "<< timer.renderToSecond() << " s." << std::endl;
  methods << "Method 3" << "\t" << timer.renderToSecond() << std::endl;

  timer.start();
  sum2=method4(MAX);
  timer.finish();
  std::cout << "Method 4: "<< sum << " -> "<< timer.renderToSecond() << " s." << std::endl;
  methods << "Method 4" << "\t" << timer.renderToSecond() << std::endl;

  timer.start();
  sum3=method5(MAX);
  timer.finish();
  std::cout << "Method 5: "<< sum << " -> "<< timer.renderToSecond() << " s." << std::endl;
  methods << "Method 5" << "\t" << timer.renderToSecond() << std::endl;

  // Compare 4 and 5 for big values
  std::cout << "Creating comparison of methods 4 and 5" << std::endl;

  pe::Timer timer4, timer5;
  for (unsigned long long upper=100; upper<= 100000000*MAX; upper *= 100) {
	timer4.start();
	sum2 = method4(upper);
	timer4.finish();
	timer5.start();
	sum3 = method5(upper);
	timer5.finish();
	std::cout << upper << "\t" << sum2 << "(" <<timer4.renderToSecond() 
	  << ")\t" << sum3 << "("<<timer5.renderToSecond() << ")"<< std::endl;
	scaling << upper << "\t" << timer4.renderToSecond() 
	  << "\t" << timer5.renderToSecond() << std::endl;

  }

  methods.close();
  scaling.close();
  return 0;
}


